import pytest
import numpy as np
import holoviews as hv

from ..utilities import load_videos
from ..motion_correction import *

dpath = "./minian/test/test_movie"

param_load_videos = {
    "pattern": "msCam[0-9].avi",
    "dtype": np.uint8,
    "downsample": dict(frame=2, height=1, width=1),
    "downsample_strategy": "subset",
}

param_estimate_shift = {
    'dim': 'frame',
    'max_sh': 20}

@pytest.fixture
def varr():
    return load_videos(dpath, **param_load_videos)


def test_estimate_shifts(varr):
    shifts = estimate_shifts(varr, **param_estimate_shift)
    assert (
        shifts.any() != varr.any()
    )
    
def test_apply_shifts(varr):
    shifts = estimate_shifts(varr, **param_estimate_shift)
    varr_ref = apply_shifts(varr, shifts)
    assert (
        varr_ref.any() != shifts.any()
    ) #If not equal, shifts have changed


import os
import warnings

import cv2
import cvxpy as cvx
import dask as da
import dask.array as darr
import networkx as nx
import numpy as np
import pyfftw.interfaces.numpy_fft as numpy_fft
import scipy.sparse
import sparse
import xarray as xr
import zarr
from scipy.linalg import lstsq, toeplitz
from scipy.ndimage import label
from scipy.signal import butter, lfilter, welch
from scipy.sparse import dia_matrix
from skimage import morphology as moph
from sklearn.linear_model import LassoLars
from sklearn.utils import parallel_backend
from statsmodels.tsa.stattools import acovf

from .utilities import rechunk_like, save_minian


def get_noise_fft(varr, noise_range=(0.25, 0.5), noise_method="logmexp"):
    """[summary]

    Args:
        varr ([type]): [description]
        noise_range (tuple, optional): [description]. Defaults to (0.25, 0.5).
        noise_method (str, optional): [description]. Defaults to 'logmexp'.

    Returns:
        [type]: [description]
    """
    sn = xr.apply_ufunc(
        _noise_fft,
        varr,
        input_core_dims=[["frame"]],
        output_core_dims=[[]],
        dask="parallelized",
        vectorize=True,
        kwargs=dict(noise_range=noise_range, noise_method=noise_method),
        output_dtypes=[np.float],
    )
    return sn


def _noise_fft(px, noise_range=(0.25, 0.5), noise_method="logmexp"):
    """[summary]

    Args:
        px ([type]): [description]
        noise_range (tuple, optional): [description]. Defaults to (0.25, 0.5).
        noise_method (str, optional): [description]. Defaults to 'logmexp'.

    Returns:
        [type]: [description]
    """
    _T = len(px)
    nr = np.around(np.array(noise_range) * _T).astype(int)
    px_band = (1 / _T * np.abs(numpy_fft.rfft(px)) ** 2)[nr[0] : nr[1]]
    if noise_method == "mean":
        return np.sqrt(px_band.mean())
    elif noise_method == "median":
        return np.sqrt(px_band.median())
    elif noise_method == "logmexp":
        eps = np.finfo(px_band.dtype).eps
        return np.sqrt(np.exp(np.log(px_band + eps).mean()))
    elif noise_method == "sum":
        return np.sqrt(px_band.sum())


def get_noise_welch(
    varr, noise_range=(0.25, 0.5), noise_method="logmexp", compute=True
):
    print("estimating noise")
    sn = xr.apply_ufunc(
        noise_welch,
        varr.chunk(dict(frame=-1)),
        input_core_dims=[["frame"]],
        dask="parallelized",
        vectorize=True,
        kwargs=dict(noise_range=noise_range, noise_method=noise_method),
        output_dtypes=[varr.dtype],
    )
    if compute:
        sn = sn.compute()
    return sn


def noise_welch(y, noise_range, noise_method):
    """[summary]

    Args:
        y ([type]): [description]
        noise_range ([type]): [description]
        noise_method ([type]): [description]

    Returns:
        [type]: [description]
    """
    ff, Pxx = welch(y)
    mask0, mask1 = ff > noise_range[0], ff < noise_range[1]
    mask = np.logical_and(mask0, mask1)
    Pxx_ind = Pxx[mask]
    sn = {
        "mean": lambda x: np.sqrt(np.mean(x / 2)),
        "median": lambda x: np.sqrt(np.median(x / 2)),
        "logmexp": lambda x: np.sqrt(np.exp(np.mean(np.log(x / 2)))),
    }[noise_method](Pxx_ind)
    return sn


def update_spatial(
    Y,
    A,
    b,
    C,
    f,
    sn,
    dl_wnd=5,
    sparse_penal=0.5,
    update_background=True,
    normalize=True,
    size_thres=(9, None),
    in_memory=False,
):
    """[summary]

    Args:
        Y ([type]): [description]
        A ([type]): [description]
        b ([type]): [description]
        C ([type]): [description]
        f ([type]): [description]
        sn ([type]): [description]
        dl_wnd (int, optional): [description]. Defaults to 5.
        sparse_penal (float, optional): [description]. Defaults to 0.5.
        update_background (bool, optional): [description]. Defaults to True.
        normalize (bool, optional): [description]. Defaults to True.
        size_thres (str, optional): [description]. Defaults to '(9, None)'.
        in_memory (bool, optional): [description]. Defaults to False.

    Returns:
        [type]: [description]
    """
    _T = Y.sizes["frame"]
    intpath = os.environ["MINIAN_INTERMEDIATE"]
    if in_memory:
        C_store = C.compute().values
    else:
        C_path = os.path.join(intpath, C.name + ".zarr", C.name)
        C_store = zarr.open_array(C_path)
    print("estimating penalty parameter")
    alpha = sparse_penal * sn
    alpha = alpha.persist()
    print("computing subsetting matrix")
    selem = moph.disk(dl_wnd)
    sub = xr.apply_ufunc(
        cv2.dilate,
        A,
        input_core_dims=[["height", "width"]],
        output_core_dims=[["height", "width"]],
        vectorize=True,
        kwargs=dict(kernel=selem),
        dask="parallelized",
        output_dtypes=[A.dtype],
    )
    sub = sub > 0
    sub = save_minian(
        sub.rename("sub"),
        intpath,
        overwrite=True,
        chunks={
            "unit_id": -1,
            "height": Y.data.chunksize[1],
            "width": Y.data.chunksize[2],
        },
    )
    sub.data = sub.data.map_blocks(sparse.COO)
    if update_background:
        b_in = rechunk_like(b > 0, Y).assign_coords(unit_id=-1).expand_dims("unit_id")
        b_in.data = b_in.data.map_blocks(sparse.COO)
        b_in = b_in.persist()
        sub = xr.concat([sub, b_in], "unit_id").chunk({"unit_id": -1})
        f_in = f.persist().data
    else:
        f_in = None
    print("fitting spatial matrix")
    A_new = update_spatial_block(
        Y.transpose("height", "width", "frame").data,
        alpha.data,
        sub.transpose("height", "width", "unit_id").data,
        C_store=C_store,
        f=f_in,
    )
    if normalize:
        A_new = A_new / A_new.sum(axis=(0, 1))
    A_new = xr.DataArray(
        darr.moveaxis(A_new, -1, 0).map_blocks(lambda a: a.todense(), dtype=A.dtype),
        dims=["unit_id", "height", "width"],
        coords={
            "unit_id": sub.coords["unit_id"],
            "height": A.coords["height"],
            "width": A.coords["width"],
        },
    )
    with parallel_backend("dask"):
        A_new = save_minian(
            A_new.rename("A_new"),
            intpath,
            overwrite=True,
            chunks={"unit_id": 1, "height": -1, "width": -1},
        )
    if update_background:
        print("updating background")
        b_new = A_new.sel(unit_id=-1)
        f_new = xr.apply_ufunc(
            darr.tensordot,
            Y,
            b_new,
            input_core_dims=[["frame", "height", "width"], ["height", "width"]],
            output_core_dims=[["frame"]],
            kwargs={"axes": [(1, 2), (0, 1)]},
            dask="allowed",
        )
        b_new = b_new.compute()
        f_new = f_new.compute()
        A_new = A_new[:-1, :, :]
    else:
        b_new = b.compute()
        f_new = f.compute()
    if size_thres:
        low, high = size_thres
        A_bin = A_new > 0
        mask = np.ones(A_new.sizes["unit_id"], dtype=bool)
        if low:
            mask = np.logical_and(
                (A_bin.sum(["height", "width"]) > low).compute(), mask
            )
        if high:
            mask = np.logical_and(
                (A_bin.sum(["height", "width"]) < high).compute(), mask
            )
        mask = xr.DataArray(
            mask, dims=["unit_id"], coords={"unit_id": A_new.coords["unit_id"].values}
        )
    else:
        mask = (A_new.sum(["height", "width"]) > 0).compute()
    print("{} out of {} units dropped".format(len(mask) - mask.sum().values, len(mask)))
    A_new = A_new.sel(unit_id=mask)
    return A_new, b_new, f_new, mask


def update_spatial_perpx(y, alpha, sub, C_store, f):
    """[summary]

    Args:
        y ([type]): [description]
        alpha ([type]): [description]
        sub ([type]): [description]
        C ([type]): [description]
        f ([type]): [description]

    Returns:
        [type]: [description]
    """
    if f is not None:
        idx = sub[:-1].nonzero()[0]
    else:
        idx = sub.nonzero()[0]
    try:
        C = C_store.get_orthogonal_selection((idx, slice(None))).T
    except AttributeError:
        C = C_store[idx, :].T
    if (f is not None) and sub[-1]:
        C = np.concatenate([C, f.reshape((-1, 1))], axis=1)
        idx = np.concatenate([idx, np.array(len(sub) - 1).reshape(-1)])
    clf = LassoLars(alpha=alpha, positive=True)
    coef = clf.fit(C, y).coef_
    mask = coef > 0
    coef = coef[mask]
    idx = idx[mask]
    return sparse.COO(coords=idx, data=coef, shape=sub.shape)


@darr.as_gufunc(signature="(f),(),(u)->(u)", output_dtypes=float)
def update_spatial_block(y, alpha, sub, **kwargs):
    C_store = kwargs.get("C_store")
    f = kwargs.get("f")
    crd_ls = []
    data_ls = []
    for h, w in zip(*sub.any(axis=-1).nonzero()):
        res = update_spatial_perpx(y[h, w, :], alpha[h, w], sub[h, w, :], C_store, f)
        crd = res.coords
        crd = np.concatenate([np.full_like(crd, h), np.full_like(crd, w), crd], axis=0)
        crd_ls.append(crd)
        data_ls.append(res.data)
    if data_ls:
        return sparse.COO(
            coords=np.concatenate(crd_ls, axis=1),
            data=np.concatenate(data_ls),
            shape=(sub.shape[0], sub.shape[1], sub.shape[2]),
        )
    else:
        return sparse.zeros((sub.shape[0], sub.shape[1], sub.shape[2]))


def compute_trace(Y, A, b, C, f, noise_freq=None):
    """[summary]

    Args:
        Y ([type]): [description]
        A ([type]): [description]
        b ([type]): [description]
        C ([type]): [description]
        f ([type]): [description]
        noise_freq ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """
    fms = Y.coords["frame"]
    uid = A.coords["unit_id"]
    Y = Y.data
    A = A.data.map_blocks(sparse.COO).rechunk(-1).persist()
    C = C.data.map_blocks(sparse.COO).T
    b = (
        b.fillna(0)
        .data.map_blocks(sparse.COO)
        .reshape((1, Y.shape[1], Y.shape[2]))
        .rechunk(-1)
    )
    f = f.fillna(0).data.reshape((-1, 1))
    AtA = darr.tensordot(A, A, axes=[(1, 2), (1, 2)])
    A_norm = (
        (1 / (A ** 2).sum(axis=(1, 2)))
        .map_blocks(
            lambda a: sparse.diagonalize(sparse.COO(a)), chunks=(A.shape[0], A.shape[0])
        )
        .compute()
    )
    B = darr.tensordot(f, b, axes=[(1), (0)])
    Y = Y - B
    YtA = darr.tensordot(Y, A, axes=[(1, 2), (1, 2)])
    YtA = darr.dot(YtA, A_norm)
    CtA = darr.dot(C, AtA)
    CtA = darr.dot(CtA, A_norm)
    YrA = (YtA - CtA + C).clip(0)
    YrA_ls = YrA.to_delayed().squeeze().tolist()
    for i in range(0, len(YrA_ls), 100):
        YrA_ls[i : i + 100] = da.compute(YrA_ls[i : i + 100])[0]
    YrA = darr.array(np.concatenate(YrA_ls, axis=0)).rechunk((-1, 1))
    YrA = xr.DataArray(
        YrA,
        dims=["frame", "unit_id"],
        coords={"frame": fms, "unit_id": uid},
    )
    if noise_freq:
        YrA = smooth_sig(YrA, noise_freq)
    return YrA.transpose("unit_id", "frame")


def update_temporal(
    A,
    C,
    b=None,
    f=None,
    Y=None,
    YrA=None,
    noise_freq=0.25,
    p=2,
    add_lag="p",
    jac_thres=0.1,
    sparse_penal=1,
    bseg=None,
    zero_thres=1e-8,
    max_iters=200,
    use_smooth=True,
    normalize=True,
    warm_start=False,
    post_scal=True,
    scs_fallback=False,
    concurrent_update=False,
):
    """[summary]

    Args:
        A ([type]): [description]
        C ([type]): [description]
        b ([type]): [description]
        f ([type]): [description]
        Y ([type]): [description]
        YrA ([type]): [description]
        noise_freq (float, optional): [description]. Defaults to 0.25.
        p ([type], optional): [description]. Defaults to 2.
        add_lag (str, optional): [description]. Defaults to 'p'.
        jac_thres (float, optional): [description]. Defaults to 0.1.
        bseg (str, optional): [description]. Defaults to None.
        zero_thres ([type], optional): [description]. Defaults to 1e-8.
        max_iters (int, optional): [description]. Defaults to 200.
        use_smooth (bool, optional): [description]. Defaults to True.
        normalize (bool, optional): [description]. Defaults to True.
        warm_start (bool, optional): [description]. Defaults to False.
        post_scal (bool, optional): [description]. Defaults to True.
        scs_fallback (bool, optional): [description]. Defaults to False.
        concurrent_update (bool, optional): [description]. Defaults to False.

    Returns:
        [type]: [description]
    """
    if YrA is None:
        YrA = compute_trace(Y, A, b, C, f).persist()
    Ymask = (YrA > 0).any("frame").compute()
    A, C, YrA = A.sel(unit_id=Ymask), C.sel(unit_id=Ymask), YrA.sel(unit_id=Ymask)
    print("grouping overlaping units")
    A_sps = (A.data.map_blocks(sparse.COO) > 0).rechunk(-1).persist()
    A_inter = darr.tensordot(
        A_sps.astype(np.float32), A_sps.astype(np.float32), axes=[(1, 2), (1, 2)]
    ).compute()
    A_usum = np.tile(A_sps.sum(axis=(1, 2)).compute().todense(), (A_sps.shape[0], 1))
    A_usum = A_usum + A_usum.T
    jac = scipy.sparse.csc_matrix(A_inter / (A_usum - A_inter) > jac_thres)
    unit_labels = label_connected(jac)
    YrA = YrA.assign_coords(unit_labels=("unit_id", unit_labels))
    if normalize:
        print("normalizing traces")
        YrA_norm = (
            YrA.groupby("unit_labels").apply(lambda a: a / a.sum(axis=1).mean())
            * YrA.sizes["frame"]
        )
    else:
        YrA_norm = YrA
    tn = (
        get_noise_fft(YrA_norm, noise_range=(noise_freq, 1))
        .compute()
        .chunk({"unit_id": 1})
    )
    if use_smooth:
        print("smoothing traces")
        YrA_smth = smooth_sig(YrA_norm, noise_freq)
        tn_smth = get_noise_fft(YrA_smth, noise_range=(noise_freq, 1))
    else:
        YrA_smth = YrA
        tn_smth = tn
    if p is None:
        print("estimating order p for each neuron")
        p = xr.apply_ufunc(
            get_p,
            YrA_smth,
            input_core_dims=[["frame"]],
            vectorize=True,
            dask="parallelized",
            output_dtypes=[int],
        ).clip(1)
        p = p.compute()
        p_max = p.max().values
    else:
        p_max = p
    print("estimating AR coefficients")
    g = (
        xr.apply_ufunc(
            get_ar_coef,
            YrA_smth,
            tn_smth,
            p,
            input_core_dims=[["frame"], [], []],
            output_core_dims=[["lag"]],
            kwargs=dict(pad=p_max, add_lag=add_lag),
            vectorize=True,
            dask="parallelized",
            output_dtypes=[float],
            output_sizes=dict(lag=p_max),
        )
        .assign_coords(lag=np.arange(1, p_max + 1))
        .compute()
        .chunk({"unit_id": 1})
    )
    print("updating temporal components")
    res_ls = []
    uid_ls = []
    if concurrent_update:
        grp_dim = "unit_labels"
        C = C.assign_coords(unit_labels=("unit_id", unit_labels))
    else:
        grp_dim = "unit_id"
    if warm_start:
        C.data = C.data.map_blocks(scipy.sparse.csr_matrix)
    for cur_YrA, cur_g, cur_tn, cur_C in zip(
        YrA_norm.groupby(grp_dim),
        g.groupby(grp_dim),
        tn.groupby(grp_dim),
        C.groupby(grp_dim),
    ):
        uid_ls.append(cur_YrA[1].coords["unit_id"].values.reshape(-1))
        cur_YrA, cur_g, cur_tn, cur_C = (
            cur_YrA[1].data.rechunk(-1),
            cur_g[1].data.rechunk(-1),
            cur_tn[1].data.rechunk(-1),
            cur_C[1].data.rechunk(-1),
        )
        if not concurrent_update:
            cur_YrA, cur_g, cur_tn, cur_C = (
                cur_YrA.reshape((1, -1)),
                cur_g.reshape((1, -1)),
                cur_tn.reshape(1),
                cur_C.reshape((1, -1)),
            )
        if cur_YrA.shape[0] > 50:
            warnings.warn(
                "{} units are scheduled to update together, "
                "which might be too demanding. "
                "Consider merging the units "
                "or changing jaccard threshold".format(cur_YrA.shape[0])
            )
        if not warm_start:
            cur_C = None
        res = update_temporal_cvxpy(
            cur_YrA,
            cur_g,
            cur_tn,
            c_last=cur_C,
            bseg=bseg,
            sparse_penal=sparse_penal,
            max_iters=max_iters,
            scs_fallback=scs_fallback,
            zero_thres=zero_thres,
        )
        res_ls.append(res)
    uids_new = np.concatenate(uid_ls)
    res_ls = da.compute(res_ls)[0]
    C_new = darr.concatenate([darr.array(r[0]) for r in res_ls], axis=0)
    S_new = darr.concatenate([darr.array(r[1]) for r in res_ls], axis=0)
    b0_new = darr.concatenate([darr.array(r[2]) for r in res_ls], axis=0)
    c0_new = darr.concatenate([darr.array(r[3]) for r in res_ls], axis=0)
    mask = (S_new.sum(axis=1) > 0).compute()
    print("{} out of {} units dropped".format(mask.size - mask.nnz, len(Ymask)))
    mask = mask.todense()
    C_new, S_new, b0_new, c0_new, g, uids_new_ma = (
        C_new[mask, :],
        S_new[mask, :],
        b0_new[mask, :],
        c0_new[mask, :],
        g[mask, :],
        uids_new[mask],
    )
    sig_new = C_new + b0_new + c0_new
    YrA_new = YrA[mask, :]
    if post_scal and len(sig_new) > 0:
        print("post-hoc scaling")
        scal = (
            lstsq_vec(sig_new.rechunk((1, -1)), YrA_new.data).compute().reshape((-1, 1))
        )
        C_new, S_new, b0_new, c0_new = (
            C_new * scal,
            S_new * scal,
            b0_new * scal,
            c0_new * scal,
        )
    C_new = xr.DataArray(
        C_new.map_blocks(lambda a: a.todense(), dtype=float),
        dims=["unit_id", "frame"],
        coords={
            "unit_id": uids_new_ma,
            "frame": YrA_new.coords["frame"],
        },
    )
    S_new = xr.DataArray(
        S_new.map_blocks(lambda a: a.todense(), dtype=float),
        dims=["unit_id", "frame"],
        coords={
            "unit_id": uids_new_ma,
            "frame": YrA_new.coords["frame"].values,
        },
    )
    b0_new = xr.DataArray(
        b0_new.map_blocks(lambda a: a.todense(), dtype=float),
        dims=["unit_id", "frame"],
        coords={
            "unit_id": uids_new_ma,
            "frame": YrA_new.coords["frame"].values,
        },
    )
    c0_new = xr.DataArray(
        c0_new.map_blocks(lambda a: a.todense(), dtype=float),
        dims=["unit_id", "frame"],
        coords={
            "unit_id": uids_new_ma,
            "frame": YrA_new.coords["frame"].values,
        },
    )
    mask = xr.DataArray(
        mask, dims=["unit_id"], coords={"unit_id": uids_new}
    ).reindex_like(Ymask, fill_value=False)
    return C_new, S_new, b0_new, c0_new, g, mask


@darr.as_gufunc(signature="(f),(f)->()", output_dtypes=float)
def lstsq_vec(a, b):
    a = a.reshape((-1, 1)).todense()
    return np.linalg.lstsq(a, b.squeeze(), rcond=-1)[0]


def get_ar_coef(y, sn, p, add_lag, pad=None):
    """[summary]

    Args:
        y ([type]): [description]
        sn ([type]): [description]
        p ([type]): [description]
        add_lag ([type]): [description]
        pad ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """
    if add_lag == "p":
        max_lag = p * 2
    else:
        max_lag = p + add_lag
    cov = acovf(y, fft=True)
    C_mat = toeplitz(cov[:max_lag], cov[:p]) - sn ** 2 * np.eye(max_lag, p)
    g = lstsq(C_mat, cov[1 : max_lag + 1])[0]
    if pad:
        res = np.zeros(pad)
        res[: len(g)] = g
        return res
    else:
        return g


def get_p(y):
    """[summary]

    Args:
        y ([type]): [description]

    Returns:
        [type]: [description]
    """
    dif = np.append(np.diff(y), 0)
    rising = dif > 0
    prd_ris, num_ris = label(rising)
    ext_prd = np.zeros(num_ris)
    for id_prd in range(num_ris):
        prd = y[prd_ris == id_prd + 1]
        ext_prd[id_prd] = prd[-1] - prd[0]
    id_max_prd = np.argmax(ext_prd)
    return np.sum(rising[prd_ris == id_max_prd + 1])


@darr.as_gufunc(
    signature="(u,f),(u,p),(u)->(u,f),(u,f),(u,f),(u,f)",
    output_dtypes=(float, float, float, float),
)
def update_temporal_cvxpy(y, g, sn, A=None, bseg=None, **kwargs):
    """[summary]

    spatial:
    (d, f), (u, p), (d), (d, u)
    (d, f), (p), (d), (d)
    trace:
    (u, f), (u, p), (u)
    (f), (p), ()

    Args:
        y ([type]): [description]
        g ([type]): [description]
        sn ([type]): [description]
        A ([type], optional): [description]. Defaults to None.
        bseg ([type], optional): [description]. Defaults to None

    Raises:
        ValueError: [description]
        ValueError: [description]
        ValueError: [description]

    Returns:
        [type]: [description]
    """
    # get_parameters
    sparse_penal = kwargs.get("sparse_penal")
    max_iters = kwargs.get("max_iters")
    use_cons = kwargs.get("use_cons", False)
    scs = kwargs.get("scs_fallback")
    c_last = kwargs.get("c_last")
    zero_thres = kwargs.get("zero_thres")
    # conform variables to generalize multiple unit case
    if y.ndim < 2:
        y = y.reshape((1, -1))
    if g.ndim < 2:
        g = g.reshape((1, -1))
    sn = np.atleast_1d(sn)
    if A is not None:
        if A.ndim < 2:
            A = A.reshape((-1, 1))
    # get count of frames and units
    _T = y.shape[-1]
    _u = g.shape[0]
    if A is not None:
        _d = A.shape[0]
    # construct G matrix and decay vector per unit
    dc_vec = np.zeros((_u, _T))
    G_ls = []
    for cur_u in range(_u):
        cur_g = g[cur_u, :]
        # construct first column and row
        cur_c = np.zeros(_T)
        cur_c[0] = 1
        cur_c[1 : len(cur_g) + 1] = -cur_g
        # update G with toeplitz matrix
        G_ls.append(
            cvx.Constant(
                dia_matrix(
                    (
                        np.tile(np.concatenate(([1], -cur_g)), (_T, 1)).T,
                        -np.arange(len(cur_g) + 1),
                    ),
                    shape=(_T, _T),
                ).tocsc()
            )
        )
        # update dc_vec
        cur_gr = np.roots(cur_c)
        dc_vec[cur_u, :] = np.max(cur_gr.real) ** np.arange(_T)
    # get noise threshold
    thres_sn = sn * np.sqrt(_T)
    # construct variables
    if bseg is not None:
        nseg = int(np.max(bseg) + 1)
        b_temp = np.zeros((nseg, _T))
        for iseg in range(nseg):
            b_temp[iseg, bseg == iseg] = 1
        b_cmp = cvx.Variable((_u, nseg))
    else:
        b_temp = np.ones((1, _T))
        b_cmp = cvx.Variable((_u, 1))
    b = b_cmp @ b_temp  # baseline fluorescence per unit
    c0 = cvx.Variable(_u)  # initial fluorescence per unit
    c = cvx.Variable((_u, _T))  # calcium trace per unit
    if c_last is not None:
        c.value = c_last
        warm_start = True
    else:
        warm_start = False
    s = cvx.vstack([G_ls[u] @ c[u, :] for u in range(_u)])  # spike train per unit
    # residual noise per unit
    if A is not None:
        sig = cvx.vstack(
            [
                (A * c)[px, :] + (A * b)[px, :] + (A * cvx.diag(c0) * dc_vec)[px, :]
                for px in range(_d)
            ]
        )
        noise = y - sig
    else:
        sig = cvx.vstack([c[u, :] + b[u, :] + c0[u] * dc_vec[u, :] for u in range(_u)])
        noise = y - sig
    noise = cvx.vstack([cvx.norm(noise[i, :], 2) for i in range(noise.shape[0])])
    # construct constraints
    cons = []
    cons.append(
        b >= np.broadcast_to(np.min(y, axis=-1).reshape((-1, 1)), y.shape)
    )  # baseline larger than minimum
    cons.append(c0 >= 0)  # initial fluorescence larger than 0
    cons.append(s >= 0)  # spike train non-negativity
    # noise constraints
    cons_noise = [noise[i] <= thres_sn[i] for i in range(thres_sn.shape[0])]
    try:
        obj = cvx.Minimize(cvx.sum(cvx.norm(s, 1, axis=1)))
        prob = cvx.Problem(obj, cons + cons_noise)
        if use_cons:
            _ = prob.solve(solver="ECOS")
        if not (prob.status == "optimal" or prob.status == "optimal_inaccurate"):
            if use_cons:
                warnings.warn("constrained version of problem infeasible")
            raise ValueError
    except (ValueError, cvx.SolverError):
        lam = sn * sparse_penal
        obj = cvx.Minimize(
            cvx.sum(cvx.sum(noise, axis=1) + cvx.multiply(lam, cvx.norm(s, 1, axis=1)))
        )
        prob = cvx.Problem(obj, cons)
        try:
            _ = prob.solve(solver="ECOS", warm_start=warm_start, max_iters=max_iters)
            if prob.status in ["infeasible", "unbounded", None]:
                raise ValueError
        except (cvx.SolverError, ValueError):
            try:
                if scs:
                    _ = prob.solve(solver="SCS", max_iters=200)
                if prob.status in ["infeasible", "unbounded", None]:
                    raise ValueError
            except (cvx.SolverError, ValueError):
                warnings.warn(
                    "problem status is {}, returning zero".format(prob.status),
                    RuntimeWarning,
                )
                return [sparse.zeros(c.shape, dtype=float)] * 4
    if not (prob.status == "optimal"):
        warnings.warn("problem solved sub-optimally", RuntimeWarning)
    c = np.where(c.value > zero_thres, c.value, 0)
    s = np.where(s.value > zero_thres, s.value, 0)
    b = np.where(b.value > zero_thres, b.value, 0)
    c0 = c0.value.reshape((-1, 1)) * dc_vec
    c0 = np.where(c0 > zero_thres, c0, 0)
    return sparse.COO(c), sparse.COO(s), sparse.COO(b), sparse.COO(c0)


def unit_merge(A, C, add_list=None, thres_corr=0.9):
    """[summary]

    Args:
        A ([type]): [description]
        C ([type]): [description]
        add_list ([type], optional): [description]. Defaults to None.
        thres_corr (float, optional): [description]. Defaults to 0.9.

    Returns:
        [type]: [description]
    """
    print("computing spatial overlap")
    A_sps = (A.data.map_blocks(sparse.COO) > 0).rechunk(-1).persist()
    A_inter = sparse.tril(
        darr.tensordot(
            A_sps.astype(np.float32), A_sps.astype(np.float32), axes=[(1, 2), (1, 2)]
        ).compute(),
        k=-1,
    )
    print("computing temporal correlation")
    corr_ls = []
    row_idx = []
    col_idx = []
    C = C - C.mean("frame")
    std = np.sqrt((C ** 2).sum("frame"))
    for i, j in zip(*A_inter.nonzero()):
        C_i, C_j, std_i, std_j = (
            C.isel(unit_id=i),
            C.isel(unit_id=j),
            std.isel(unit_id=i),
            std.isel(unit_id=j),
        )
        corr = (C_i * C_j).sum() / (std_i * std_j)
        corr_ls.append(corr)
        row_idx.append(i)
        col_idx.append(j)
    corr_ls = da.compute(corr_ls)[0]
    print("labeling units to be merged")
    adj = (
        scipy.sparse.csr_matrix((corr_ls, (row_idx, col_idx)), shape=A_inter.shape)
        > thres_corr
    )
    adj = adj + adj.T
    unit_labels = xr.apply_ufunc(
        label_connected,
        adj,
        input_core_dims=[["unit_id", "unit_id_cp"]],
        output_core_dims=[["unit_id"]],
    )
    print("merging units")
    A_merge = (
        A.assign_coords(unit_labels=("unit_id", unit_labels))
        .groupby("unit_labels")
        .mean("unit_id")
        .rename(unit_labels="unit_id")
    )
    C_merge = (
        C.assign_coords(unit_labels=("unit_id", unit_labels))
        .groupby("unit_labels")
        .mean("unit_id")
        .rename(unit_labels="unit_id")
    )
    if add_list:
        for ivar, var in enumerate(add_list):
            var_mrg = (
                var.assign_coords(unit_labels=("unit_id", unit_labels))
                .groupby("unit_labels")
                .mean("unit_id")
                .rename(unit_labels="unit_id")
            )
            add_list[ivar] = var_mrg
        return A_merge, C_merge, add_list
    else:
        return A_merge, C_merge


def label_connected(adj, only_connected=False):
    """[summary]

    Args:
        adj ([type]): [description]
        only_connected (bool, optional): [description]. Defaults to False.

    Returns:
        [type]: [description]
    """
    try:
        np.fill_diagonal(adj, 0)
        adj = np.triu(adj)
        g = nx.convert_matrix.from_numpy_matrix(adj)
    except:
        g = nx.convert_matrix.from_scipy_sparse_matrix(adj)
    labels = np.zeros(adj.shape[0], dtype=np.int)
    for icomp, comp in enumerate(nx.connected_components(g)):
        comp = list(comp)
        if only_connected and len(comp) == 1:
            labels[comp] = -1
        else:
            labels[comp] = icomp
    return labels


def smooth_sig(sig, freq, method="fft", btype="low"):
    """[summary]

    Args:
        sig ([type]): [description]
        freq ([type]): [description]
        method ([type]): [description]. Defaults to 'fft'
        btype (str, optional): [description]. Defaults to 'low'.

    Returns:
        [type]: [description]
    """
    if method == "fft":
        _T = sig.sizes["frame"]
        if btype == "low":
            zero_range = slice(int(freq * _T), None)
        elif btype == "high":
            zero_range = slice(None, int(freq * _T))

        def filt_func(x):
            xfft = np.fft.rfft(x)
            xfft[zero_range] = 0
            return np.fft.irfft(xfft, len(x))

    elif method == "butter":
        but_b, but_a = butter(2, freq * 2, btype=btype, analog=False)
        filt_func = lambda x: lfilter(but_b, but_a, x)
    else:
        raise NotImplementedError(method)
    sig_smth = xr.apply_ufunc(
        filt_func,
        sig,
        input_core_dims=[["frame"]],
        output_core_dims=[["frame"]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[sig.dtype],
    )
    return sig_smth


def compute_AtC(A, C):
    fm, h, w = (
        C.coords["frame"].values,
        A.coords["height"].values,
        A.coords["width"].values,
    )
    A = A.data.map_blocks(sparse.COO, dtype=A.dtype).rechunk(-1)
    C = C.transpose("frame", "unit_id").data.map_blocks(sparse.COO, dtype=C.dtype)
    AtC = darr.tensordot(C, A, axes=(1, 0)).map_blocks(
        lambda a: a.todense(), dtype=A.dtype
    )
    return xr.DataArray(
        AtC,
        dims=["frame", "height", "width"],
        coords={"frame": fm, "height": h, "width": w},
    )



import itertools as itt
import os

import cv2
import dask
import dask.array as da
import numpy as np
import pandas as pd
import scipy.sparse
import sparse
import xarray as xr
from dask.delayed import delayed
from scipy.ndimage.filters import median_filter
from scipy.ndimage.measurements import label
from scipy.signal import butter, lfilter
from scipy.stats import kstest, zscore
from skimage.morphology import disk
from sklearn.mixture import GaussianMixture
from sklearn.neighbors import kneighbors_graph

from .cnmf import label_connected, smooth_sig
from .utilities import save_minian


def seeds_init(
    varr,
    wnd_size=500,
    method="rolling",
    stp_size=200,
    nchunk=100,
    max_wnd=10,
    diff_thres=2,
):
    """[summary]

    Args:
        varr ([type]): [description]
        wnd_size (int, optional): [description]. Defaults to 500.
        method (str, optional): [description]. Defaults to 'rolling'.
        stp_size (int, optional): [description]. Defaults to 200.
        nchunk (int, optional): [description]. Defaults to 100.
        max_wnd (int, optional): [description]. Defaults to 10.
        diff_thres (int, optional): [description]. Defaults to 2.

    Returns:
        [type]: [description]
    """
    int_path = os.environ["MINIAN_INTERMEDIATE"]
    print("constructing chunks")
    idx_fm = varr.coords["frame"]
    nfm = len(idx_fm)
    if method == "rolling":
        nstp = np.ceil(nfm / stp_size) + 1
        centers = np.linspace(0, nfm - 1, int(nstp))
        hwnd = np.ceil(wnd_size / 2)
        max_idx = list(
            map(
                lambda c: slice(
                    int(np.floor(c - hwnd).clip(0)), int(np.ceil(c + hwnd))
                ),
                centers,
            )
        )
    elif method == "random":
        max_idx = [np.random.randint(0, nfm - 1, wnd_size) for _ in range(nchunk)]
    print("computing max projections")
    res = [max_proj_frame(varr, cur_idx) for cur_idx in max_idx]
    max_res = xr.concat(res, "sample")
    max_res = save_minian(max_res.rename("max_res"), int_path, overwrite=True)
    print("calculating local maximum")
    loc_max = xr.apply_ufunc(
        local_max_roll,
        max_res,
        input_core_dims=[["height", "width"]],
        output_core_dims=[["height", "width"]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[np.uint8],
        kwargs=dict(k0=2, k1=max_wnd, diff=diff_thres),
    ).sum("sample")
    seeds = (
        loc_max.where(loc_max > 0).rename("seeds").to_dataframe().dropna().reset_index()
    )
    return seeds[["height", "width", "seeds"]]


def max_proj_frame(varr, idx):
    """[summary]

    Args:
        varr ([type]): [description]
        idx ([type]): [description]

    Returns:
        [type]: [description]
    """
    return varr.isel(frame=idx).max("frame")


def local_max_roll(fm, k0, k1, diff):
    """[summary]

    Args:
        fm ([type]): [description]
        k0 ([type]): [description]
        k1 ([type]): [description]
        diff ([type]): [description]

    Returns:
        [type]: [description]
    """
    max_ls = []
    for ksize in range(k0, k1):
        selem = disk(ksize)
        fm_max = local_max(fm, selem, diff)
        max_ls.append(fm_max)
    lmax = (np.stack(max_ls, axis=0).sum(axis=0) > 0).astype(np.uint8)
    nlab, max_lab = cv2.connectedComponents(lmax)
    max_res = np.zeros_like(lmax)
    for lb in range(1, nlab):
        area = max_lab == lb
        if np.sum(area) > 1:
            crds = tuple(int(np.median(c)) for c in np.where(area))
            max_res[crds] = 1
        else:
            max_res[np.where(area)] = 1
    return max_res


def local_max(fm, k, diff=0):
    """[summary]

    Args:
        fm ([type]): [description]
        k ([type]): [description]
        diff (int, optional): [description]. Defaults to 0.

    Returns:
        [type]: [description]
    """
    fm_max = cv2.dilate(fm, k)
    fm_min = cv2.erode(fm, k)
    fm_diff = ((fm_max - fm_min) > diff).astype(np.uint8)
    fm_max = (fm == fm_max).astype(np.uint8)
    return cv2.bitwise_and(fm_max, fm_diff).astype(np.uint8)


def gmm_refine(
    varr, seeds, q=(0.1, 99.9), n_components=2, valid_components=1, mean_mask=True
):
    """[summary]

    Args:
        varr ([type]): [description]
        seeds ([type]): [description]
        q (tuple, optional): [description]. Defaults to (0.1, 99.9).
        n_components (int, optional): [description]. Defaults to 2.
        valid_components (int, optional): [description]. Defaults to 1.
        mean_mask (bool, optional): [description]. Defaults to True.

    Returns:
        [type]: [description]
    """
    print("selecting seeds")
    varr_sub = varr.sel(spatial=[tuple(hw) for hw in seeds[["height", "width"]].values])
    print("computing peak-valley values")
    varr_valley = xr.apply_ufunc(
        np.percentile,
        varr_sub.chunk(dict(frame=-1)),
        input_core_dims=[["frame"]],
        kwargs=dict(q=q[0], axis=-1),
        dask="parallelized",
        output_dtypes=[varr_sub.dtype],
    )
    varr_peak = xr.apply_ufunc(
        np.percentile,
        varr_sub.chunk(dict(frame=-1)),
        input_core_dims=[["frame"]],
        kwargs=dict(q=q[1], axis=-1),
        dask="parallelized",
        output_dtypes=[varr_sub.dtype],
    )
    varr_pv = varr_peak - varr_valley
    varr_pv = varr_pv.compute()
    print("fitting GMM models")
    dat = varr_pv.values.reshape(-1, 1)
    gmm = GaussianMixture(n_components=n_components)
    gmm.fit(dat)
    idg = np.argsort(gmm.means_.reshape(-1))[-valid_components:]
    idx_valid = np.isin(gmm.predict(dat), idg)
    if mean_mask:
        idx_mean = dat > np.sort(gmm.means_)[0]
        idx_valid = np.logical_and(idx_mean.squeeze(), idx_valid)
    seeds["mask_gmm"] = idx_valid
    return seeds, varr_pv, gmm


def pnr_refine(varr, seeds, noise_freq=0.25, thres=1.5, q=(0.1, 99.9), med_wnd=None):
    """[summary]

    Args:
        varr ([type]): [description]
        seeds ([type]): [description]
        noise_freq (float, optional): [description]. Defaults to 0.25.
        thres (float, optional): [description]. Defaults to 1.5.
        q (tuple, optional): [description]. Defaults to (0.1, 99.9).
        med_wnd ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """
    print("selecting seeds")
    varr_sub = xr.concat(
        [varr.sel(height=h, width=w) for h, w in seeds[["height", "width"]].values],
        "index",
    ).assign_coords({"index": seeds.index.values})
    if med_wnd:
        print("removing baseline")
        varr = xr.apply_ufunc(
            med_baseline,
            varr_sub,
            input_core_dims=[["frame"]],
            output_core_dims=[["frame"]],
            dask="parallelized",
            kwargs={"wnd": med_wnd},
            vectorize=True,
            output_dtypes=[varr.dtype],
        )
    print("computing peak-noise ratio")
    pnr = xr.apply_ufunc(
        pnr_perseed,
        varr_sub,
        input_core_dims=[["frame"]],
        output_core_dims=[[]],
        kwargs={"freq": noise_freq, "q": q},
        vectorize=True,
        dask="parallelized",
        output_dtypes=[float],
    ).compute()
    if thres == "auto":
        gmm = GaussianMixture(n_components=2)
        gmm.fit(np.nan_to_num(pnr.values.reshape(-1, 1)))
        idg = np.argsort(gmm.means_.reshape(-1))[-1]
        idx_valid = np.isin(gmm.predict(pnr.values.reshape(-1, 1)), idg)
        seeds["mask_pnr"] = idx_valid
    else:
        mask = pnr > thres
        mask_df = mask.to_pandas().rename("mask_pnr")
        seeds["mask_pnr"] = mask_df
        gmm = None
    return seeds, pnr, gmm


def ptp_q(a, q):
    return np.percentile(a, q[1]) - np.percentile(a, q[0])


def pnr_perseed(a, freq, q):
    ptp = ptp_q(a, q)
    but_b, but_a = butter(2, freq, btype="high", analog=False)
    a = lfilter(but_b, but_a, a).real
    ptp_noise = ptp_q(a, q)
    return ptp / ptp_noise


def med_baseline(a, wnd):
    base = median_filter(a, size=wnd)
    a -= base
    return a


def intensity_refine(varr, seeds, thres_mul=2):
    """[summary]

    Args:
        varr ([type]): [description]
        seeds ([type]): [description]
        thres_mul (int, optional): [description]. Defaults to 2.

    Returns:
        [type]: [description]
    """
    try:
        fm_max = varr.max("frame")
    except ValueError:
        print("using input as max projection")
        fm_max = varr
    bins = np.around(fm_max.sizes["height"] * fm_max.sizes["width"] / 10).astype(int)
    hist, edges = np.histogram(fm_max, bins=bins)
    try:
        thres = edges[int(np.around(np.argmax(hist) * thres_mul))]
    except IndexError:
        print("threshold out of bound, returning input")
        return seeds
    mask = (fm_max > thres).stack(spatial=["height", "width"])
    mask_df = mask.to_pandas().rename("mask_int").reset_index()
    seeds = pd.merge(seeds, mask_df, on=["height", "width"], how="left")
    return seeds


def ks_refine(varr, seeds, sig=0.01):
    """[summary]

    Args:
        varr ([type]): [description]
        seeds ([type]): [description]
        sig (float, optional): [description]. Defaults to 0.01.

    Returns:
        [type]: [description]
    """
    print("selecting seeds")
    varr_sub = xr.concat(
        [varr.sel(height=h, width=w) for h, w in seeds[["height", "width"]].values],
        "index",
    ).assign_coords({"index": seeds.index.values})
    print("performing KS test")
    ks = xr.apply_ufunc(
        ks_perseed,
        varr_sub,
        input_core_dims=[["frame"]],
        output_core_dims=[[]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[float],
    ).compute()
    ks = (ks < sig).to_pandas().rename("mask_ks")
    seeds["mask_ks"] = ks
    return seeds


def ks_perseed(a):
    a = zscore(a)
    return kstest(a, "norm")[1]


def seeds_merge(varr, max_proj, seeds, thres_dist=5, thres_corr=0.6, noise_freq=None):
    """[summary]

    Args:
        varr ([type]): [description]
        seeds ([type]): [description]
        thres_dist (int, optional): [description]. Defaults to 5.
        thres_corr (float, optional): [description]. Defaults to 0.6.
        noise_freq (str, optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """
    if noise_freq:
        varr = smooth_sig(varr, noise_freq)
    print("computing distance")
    dist = kneighbors_graph(seeds[["height", "width"]], n_neighbors=1, mode="distance")
    print("computing correlations")
    corr_ls = []
    row_idx = []
    col_idx = []
    varr = varr - varr.mean("frame")
    std = np.sqrt((varr ** 2).sum("frame"))
    for i, j in zip(*dist.nonzero()):
        if dist[i, j] < thres_dist:
            hi, hj, wi, wj = (
                seeds.iloc[i]["height"],
                seeds.iloc[j]["height"],
                seeds.iloc[i]["width"],
                seeds.iloc[j]["width"],
            )
            varr_i, varr_j, std_i, std_j = (
                varr.sel(height=hi, width=wi),
                varr.sel(height=hj, width=wj),
                std.sel(height=hi, width=wi),
                std.sel(height=hj, width=wj),
            )
            corr = (varr_i * varr_j).sum() / (std_i * std_j)
            corr_ls.append(corr)
            row_idx.append(i)
            col_idx.append(j)
    corr_ls = dask.compute(corr_ls)[0]
    print("merging seeds")
    adj = (
        scipy.sparse.csr_matrix((corr_ls, (row_idx, col_idx)), shape=dist.shape)
        > thres_corr
    )
    adj = adj + adj.T
    labels = label_connected(adj, only_connected=True)
    iso = np.where(labels < 0)[0]
    seeds_final = set(iso.tolist())
    for cur_cmp in np.unique(labels):
        if cur_cmp < 0:
            continue
        cur_smp = np.where(labels == cur_cmp)[0]
        cur_max = np.array(
            [
                max_proj.sel(
                    height=seeds.iloc[s]["height"], width=seeds.iloc[s]["width"]
                )
                for s in cur_smp
            ]
        )
        max_seed = cur_smp[np.argmax(cur_max)]
        seeds_final.add(max_seed)
    seeds["mask_mrg"] = False
    seeds.loc[list(seeds_final), "mask_mrg"] = True
    return seeds


def initA(varr, seeds, thres_corr=0.8, wnd=10, noise_freq=None):
    seeds = seeds.sort_values(["height", "width"])
    if noise_freq:
        print("smoothing signal")
        varr = smooth_sig(varr, noise_freq)
    print("computing correlations")
    varr = varr - varr.mean("frame")
    std = np.sqrt((varr ** 2).sum("frame"))
    res_ls = [
        initA_perseed(varr, std, h, w, wnd, thres_corr)
        for h, w in zip(seeds["height"].values, seeds["width"].values)
    ]
    A_ls = []
    for i in range(0, len(res_ls), 50):
        cur_ls = dask.compute(res_ls[i : i + 50])[0]
        A_ls.extend([a.chunk() for a in cur_ls])
    A = xr.concat(A_ls, "unit_id")
    A = A.assign_coords(unit_id=np.arange(A.sizes["unit_id"]))
    A.data = A.data.map_blocks(lambda a: a.todense(), dtype=float)
    return A


def initA_perseed(varr, std, h, w, wnd, thres_corr):
    ih = np.where(varr.coords["height"] == h)[0][0]
    iw = np.where(varr.coords["width"] == w)[0][0]
    h_sur, w_sur = (
        np.arange(max(ih - wnd, 0), min(ih + wnd, varr.sizes["height"])),
        np.arange(max(iw - wnd, 0), min(iw + wnd, varr.sizes["width"])),
    )
    sur = varr.isel(height=h_sur, width=w_sur)
    corr = (
        varr.isel(height=ih, width=iw).dot(sur) / std.isel(height=ih, width=iw) / std
    ).data.reshape(-1)
    corr = da.where(corr > thres_corr, corr, 0)
    crds = np.array(list(itt.product(h_sur, w_sur))).T
    corr = delayed(sparse.COO)(
        crds,
        corr,
        shape=(varr.sizes["height"], varr.sizes["width"]),
    )
    corr = da.from_delayed(
        corr, shape=(varr.sizes["height"], varr.sizes["width"]), dtype=float
    )
    return xr.DataArray(
        corr,
        dims=["height", "width"],
        coords={
            "height": varr.coords["height"].values,
            "width": varr.coords["width"].values,
        },
    )


def initC(varr, A):
    uids = A.coords["unit_id"]
    fms = varr.coords["frame"]
    A = A.data.map_blocks(sparse.COO).map_blocks(lambda a: a / a.sum()).rechunk(-1)
    C = da.tensordot(A, varr, axes=[(1, 2), (1, 2)])
    C = xr.DataArray(
        C, dims=["unit_id", "frame"], coords={"unit_id": uids, "frame": fms}
    )
    return C


def initbf(varr, A, C):
    A = A.data.map_blocks(sparse.COO).compute()
    Yb = (varr - da.tensordot(C, A, axes=[(0,), (0,)])).clip(0)
    intpath = os.environ["MINIAN_INTERMEDIATE"]
    Yb = save_minian(Yb.rename("Yb"), intpath, overwrite=True)
    b = Yb.mean("frame")
    f = Yb.mean(["height", "width"])
    return b, f


@da.as_gufunc(signature="(h, w)->(h, w)", output_dtypes=int, allow_rechunk=True)
def da_label(im):
    """[summary]

    Args:
        im ([type]): [description]

    Returns:
        [type]: [description]
    """
    return label(im)[0]



import os
import re
import shutil
import warnings
from copy import deepcopy
from os import listdir
from os.path import isdir
from os.path import join as pjoin
from pathlib import Path
from uuid import uuid4

import cv2
import dask as da
import dask.array as darr
import numpy as np
import pandas as pd
import psutil
import rechunker
import xarray as xr
import zarr as zr
from natsort import natsorted
from tifffile import TiffFile, imread


def load_videos(
    vpath,
    pattern=r"msCam[0-9]+\.avi$",
    dtype=np.float64,
    in_memory=False,
    downsample=None,
    downsample_strategy="subset",
    post_process=None,
):
    """Load videos from a folder.

    Load videos from the folder specified in `vpath` and according to the regex
    `pattern`, then concatenate them together across time and return a
    `xarray.DataArray` representation of the concatenated videos. The default
    assumption is video filenames start with ``msCam`` followed by at least a
    number, and then followed by ``.avi``. In addition, it is assumed that the
    name of the folder correspond to a recording session identifier.

    Parameters
    ----------
    vpath : str
        The path to search for videos
    pattern : str, optional
        The pattern that describes filenames of videos. (Default value =
        'msCam[0-9]+\.avi')

    Returns
    -------
    xarray.DataArray or None
        The labeled 3-d array representation of the videos with dimensions:
        ``frame``, ``height`` and ``width``. Returns ``None`` if no data was
        found in the specified folder.

    """
    vpath = os.path.normpath(vpath)
    ssname = os.path.basename(vpath)
    vlist = natsorted(
        [vpath + os.sep + v for v in os.listdir(vpath) if re.search(pattern, v)]
    )
    if not vlist:
        raise FileNotFoundError(
            "No data with pattern {}"
            " found in the specified folder {}".format(pattern, vpath)
        )
    print("loading {} videos in folder {}".format(len(vlist), vpath))

    file_extension = os.path.splitext(vlist[0])[1]
    if file_extension in (".avi", ".mkv"):
        movie_load_func = load_avi_lazy
    elif file_extension == ".tif":
        movie_load_func = load_tif_lazy
    else:
        raise ValueError("Extension not supported.")

    varr_list = [movie_load_func(v) for v in vlist]
    varr = darr.concatenate(varr_list, axis=0)
    varr = xr.DataArray(
        varr,
        dims=["frame", "height", "width"],
        coords=dict(
            frame=np.arange(varr.shape[0]),
            height=np.arange(varr.shape[1]),
            width=np.arange(varr.shape[2]),
        ),
    )
    if dtype:
        varr = varr.astype(dtype)
    if downsample:
        bin_eg = {d: np.arange(0, varr.sizes[d], w) for d, w in downsample.items()}
        if downsample_strategy == "mean":
            varr = (
                varr.coarsen(**downsample, boundary="trim")
                .mean()
                .assign_coords(**bin_eg)
            )
        elif downsample_strategy == "subset":
            varr = varr.sel(**bin_eg)
        else:
            warnings.warn("unrecognized downsampling strategy", RuntimeWarning)
    varr = varr.rename("fluorescence")
    if post_process:
        varr = post_process(varr, vpath, ssname, vlist, varr_list)
    return varr


def load_tif_lazy(fname):
    """[summary]

    Args:
        fname ([type]): [description]

    Returns:
        [type]: [description]
    """
    data = TiffFile(fname)
    f = len(data.pages)

    fmread = da.delayed(load_tif_perframe)
    flist = [fmread(fname, i) for i in range(f)]

    sample = flist[0].compute()
    arr = [
        da.array.from_delayed(fm, dtype=sample.dtype, shape=sample.shape)
        for fm in flist
    ]
    return da.array.stack(arr, axis=0)


def load_tif_perframe(fname, fid):
    """[summary]

    Args:
        fname ([type]): [description]
        fid ([type]): [description]

    Returns:
        [type]: [description]
    """
    return imread(fname, key=fid)


def load_avi_lazy(fname):
    """[summary]

    Args:
        fname ([type]): [description]

    Returns:
        [type]: [description]
    """
    cap = cv2.VideoCapture(fname)
    f = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    fmread = da.delayed(load_avi_perframe)
    flist = [fmread(fname, i) for i in range(f)]
    sample = flist[0].compute()
    arr = [
        da.array.from_delayed(fm, dtype=sample.dtype, shape=sample.shape)
        for fm in flist
    ]
    return da.array.stack(arr, axis=0)


def load_avi_perframe(fname, fid):
    """[summary]

    Args:
        fname ([type]): [description]
        fid ([type]): [description]

    Returns:
        [type]: [description]
    """
    cap = cv2.VideoCapture(fname)
    h = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    w = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    cap.set(cv2.CAP_PROP_POS_FRAMES, fid)
    ret, fm = cap.read()
    if ret:
        return np.flip(cv2.cvtColor(fm, cv2.COLOR_RGB2GRAY), axis=0)
    else:
        print("frame read failed for frame {}".format(fid))
        return np.zeros((h, w))


def open_minian(dpath, var_name, return_dict=False):
    dpath = os.path.normpath(dpath)
    fp = os.path.join(dpath, var_name + ".zarr")
    return xr.open_zarr(fp)[var_name]


def open_minian_mf(
    dpath,
    index_dims,
    result_format="xarray",
    pattern=r"minian\.[0-9]+$",
    sub_dirs=[],
    exclude=True,
    **kwargs
):
    """[summary]

    Args:
        dpath ([type]): [description]
        index_dims ([type]): [description]
        result_format (str, optional): [description]. Defaults to 'xarray'.
        pattern (regexp, optional): [description]. Defaults to r'minian\.[0-9]+$'.
        sub_dirs (list, optional): [description]. Defaults to [].
        exclude (bool, optional): [description]. Defaults to True.

    Raises:
        NotImplementedError: [description]

    Returns:
        [type]: [description]
    """
    minian_dict = dict()
    for nextdir, dirlist, filelist in os.walk(dpath, topdown=False):
        nextdir = os.path.abspath(nextdir)
        cur_path = Path(nextdir)
        dir_tag = bool(
            (
                (any([Path(epath) in cur_path.parents for epath in sub_dirs]))
                or nextdir in sub_dirs
            )
        )
        if exclude == dir_tag:
            continue
        flist = list(filter(lambda f: re.search(pattern, f), filelist + dirlist))
        if flist:
            print("opening dataset under {}".format(nextdir))
            if len(flist) > 1:
                warnings.warn("multiple dataset found: {}".format(flist))
            fname = flist[-1]
            print("opening {}".format(fname))
            minian = open_minian(nextdir, fname=fname, **kwargs)
            key = tuple([np.array_str(minian[d].values) for d in index_dims])
            minian_dict[key] = minian
            print(["{}: {}".format(d, v) for d, v in zip(index_dims, key)])

    if result_format == "xarray":
        return xrconcat_recursive(minian_dict, index_dims)
    elif result_format == "pandas":
        minian_df = pd.Series(minian_dict).rename("minian")
        minian_df.index.set_names(index_dims, inplace=True)
        return minian_df.to_frame()
    else:
        raise NotImplementedError("format {} not understood".format(result_format))


def save_minian(
    var, dpath, meta_dict=None, overwrite=False, chunks=None, mem_limit="200MB"
):
    """[summary]

    Args:
        var ([type]): [description]
        dpath ([type]): [description]
        fname (str, optional): [description]. Defaults to 'minian'.
        backend (str, optional): [description]. Defaults to 'netcdf'.
        meta_dict ([type], optional): [description]. Defaults to None.
        overwrite (bool, optional): [description]. Defaults to False.

    Raises:
        NotImplementedError: [description]

    Returns:
        [type]: [description]
    """
    dpath = os.path.normpath(dpath)
    Path(dpath).mkdir(parents=True, exist_ok=True)
    ds = var.to_dataset()
    if meta_dict is not None:
        pathlist = os.path.abspath(dpath).split(os.sep)
        ds = ds.assign_coords(
            **dict([(dn, pathlist[di]) for dn, di in meta_dict.items()])
        )
    md = {True: "a", False: "w-"}[overwrite]
    fp = os.path.join(dpath, var.name + ".zarr")
    if overwrite:
        try:
            shutil.rmtree(fp)
        except FileNotFoundError:
            pass
    ds.to_zarr(fp, mode=md)
    if chunks is not None:
        chunks = {d: var.sizes[d] if v <= 0 else v for d, v in chunks.items()}
        dst_path = os.path.join(dpath, str(uuid4()))
        temp_path = os.path.join(dpath, str(uuid4()))
        zstore = zr.open(fp)
        rechk = rechunker.rechunk(
            zstore[var.name], chunks, mem_limit, dst_path, temp_store=temp_path
        )
        rechk.execute()
        try:
            shutil.rmtree(temp_path)
        except FileNotFoundError:
            pass
        arr_path = os.path.join(fp, var.name)
        for f in os.listdir(arr_path):
            os.remove(os.path.join(arr_path, f))
        for f in os.listdir(dst_path):
            os.rename(os.path.join(dst_path, f), os.path.join(arr_path, f))
        os.rmdir(dst_path)
    return open_minian(dpath, var.name)


def xrconcat_recursive(var, dims):
    """[summary]

    Args:
        var ([type]): [description]
        dims ([type]): [description]

    Raises:
        NotImplementedError: [description]

    Returns:
        [type]: [description]
    """
    if len(dims) > 1:
        if type(var) is dict:
            var_dict = var
        elif type(var) is list:
            var_dict = {tuple([np.asscalar(v[d]) for d in dims]): v for v in var}
        else:
            raise NotImplementedError("type {} not supported".format(type(var)))
        try:
            var_dict = {k: v.to_dataset() for k, v in var_dict.items()}
        except AttributeError:
            pass
        var_ps = pd.Series(var_dict)
        var_ps.index.set_names(dims, inplace=True)
        xr_ls = []
        for idx, v in var_ps.groupby(level=dims[0]):
            v.index = v.index.droplevel(dims[0])
            xarr = xrconcat_recursive(v.to_dict(), dims[1:])
            xr_ls.append(xarr)
        return xr.concat(xr_ls, dim=dims[0])
    else:
        if type(var) is dict:
            var = var.values()
        return xr.concat(var, dim=dims[0])


def update_meta(dpath, pattern=r"^minian\.nc$", meta_dict=None, backend="netcdf"):
    """[summary]

    Args:
        dpath ([type]): [description]
        pattern (regexp, optional): [description]. Defaults to r'^minian\.nc$'.
        meta_dict ([type], optional): [description]. Defaults to None.
        backend (str, optional): [description]. Defaults to 'netcdf'.

    Raises:
        NotImplementedError: [description]
    """
    for dirpath, dirnames, fnames in os.walk(dpath):
        if backend == "netcdf":
            fnames = filter(lambda fn: re.search(pattern, fn), fnames)
        elif backend == "zarr":
            fnames = filter(lambda fn: re.search(pattern, fn), dirnames)
        else:
            raise NotImplementedError("backend {} not supported".format(backend))
        for fname in fnames:
            f_path = os.path.join(dirpath, fname)
            pathlist = os.path.normpath(dirpath).split(os.sep)
            new_ds = xr.Dataset()
            old_ds = open_minian(f_path, f_path, backend)
            new_ds.attrs = deepcopy(old_ds.attrs)
            old_ds.close()
            new_ds = new_ds.assign_coords(
                **dict(
                    [(cdname, pathlist[cdval]) for cdname, cdval in meta_dict.items()]
                )
            )
            if backend == "netcdf":
                new_ds.to_netcdf(f_path, mode="a")
            elif backend == "zarr":
                new_ds.to_zarr(f_path, mode="w")
            print("updated: {}".format(f_path))


def get_chk(arr):
    """[summary]

    Args:
        arr ([type]): [description]

    Returns:
        [type]: [description]
    """
    return {d: c for d, c in zip(arr.dims, arr.chunks)}


def rechunk_like(x, y):
    """[summary]

    Args:
        x ([type]): [description]
        y ([type]): [description]

    Returns:
        [type]: [description]
    """
    try:
        dst_chk = get_chk(y)
        comm_dim = set(x.dims).intersection(set(dst_chk.keys()))
        dst_chk = {d: max(dst_chk[d]) for d in comm_dim}
        return x.chunk(dst_chk)
    except TypeError:
        return x.compute()


def get_optimal_chk(ref, arr=None, dim_grp=None, ncores="auto", mem_limit="auto"):
    """[summary]

    Args:
        ref ([type]): [description]
        arr ([type], optional): [description]. Defaults to None.
        dim_grp ([type], optional): [description]. Defaults to None.
        ncores (str, optional): [description]. Defaults to 'auto'.
        mem_limit (str, optional): [description]. Defaults to 'auto'.

    Returns:
        [type]: [description]
    """
    if arr is None:
        arr = ref
    szs = ref.sizes
    if ncores == "auto":
        ncores = psutil.cpu_count()
    if mem_limit == "auto":
        mem_limit = psutil.virtual_memory().available / (1024 ** 2)
    tempsz = (
        1000
        * (3 * szs["height"] * szs["width"] + 7 * szs["frame"])
        * ref.dtype.itemsize
        / (1024 ** 2)
    )
    csize = int(np.floor((mem_limit - tempsz) / ncores / 4))
    if csize < 64:
        warnings.warn(
            "estimated memory limit is smaller than 64MiB. Using 64MiB chunksize instead. "
        )
        csize = 64
    csize = int(np.floor((mem_limit) / ncores))
    if csize > 512:
        warnings.warn(
            "estimated memory limit is bigger than 512MiB. Using 512MiB chunksize instead. "
        )
        csize = 512
    dims = arr.dims
    if not dim_grp:
        dim_grp = [(d,) for d in dims]
    opt_chk = dict()
    for dg in dim_grp:
        d_rest = set(dims) - set(dg)
        dg_dict = {d: "auto" for d in dg}
        dr_dict = {d: -1 for d in d_rest}
        dg_dict.update(dr_dict)
        with da.config.set({"array.chunk-size": "{}MiB".format(csize)}):
            arr_chk = arr.chunk(dg_dict)
        re_dict = {d: c for d, c in zip(dims, arr_chk.chunks)}
        re_dict = {d: max(re_dict[d]) for d in dg}
        opt_chk.update(re_dict)
    return opt_chk



import pytest

import itertools as itt
import os
import sys
import ffmpeg
import shutil

import holoviews as hv
import numpy as np
import xarray as xr
from dask.distributed import Client, LocalCluster
from holoviews.operation.datashader import datashade
from holoviews.util import Dynamic
from IPython.core.display import display

# Set up Initial Basic Parameters#
minian_path = "."
dpath = "./minian/test/test_movie"
dpath_fixture = "./minian/test/test_movie_fixture"
minian_ds_path = os.path.join(dpath, "minian")
intpath = "./minian/test/minian_intermediate"
subset = dict(frame=slice(0, None))
subset_mc = None
interactive = True
output_size = 100
param_save_minian = {
    "dpath": minian_ds_path,
    "meta_dict": dict(session_id=-1, session=-2, animal=-3),
    "overwrite": True,
}

# Pre-processing Parameters#
param_load_videos = {
    "pattern": "msCam[0-9]+\.avi$",
    "dtype": np.uint8,
    "downsample": dict(frame=1, height=1, width=1),
    "downsample_strategy": "subset",
}
param_denoise = {"method": "median", "ksize": 7}
param_background_removal = {"method": "tophat", "wnd": 15}

# Motion Correction Parameters#
subset_mc = None
param_estimate_shift = {"dim": "frame", "max_sh": 20}

# Initialization Parameters#
param_seeds_init = {
    "wnd_size": 1000,
    "method": "rolling",
    "stp_size": 500,
    "nchunk": 100,
    "max_wnd": 15,
    "diff_thres": 2,
}
param_pnr_refine = {"noise_freq": 0.06, "thres": 1, "med_wnd": None}
param_ks_refine = {"sig": 0.05}
param_seeds_merge = {"thres_dist": 10, "thres_corr": 0.7, "noise_freq": 0.06}
param_initialize = {"thres_corr": 0.8, "wnd": 15, "noise_freq": 0.06}

# CNMF Parameters#
param_get_noise = {"noise_range": (0.06, 0.5), "noise_method": "logmexp"}
param_first_spatial = {
    "dl_wnd": 5,
    "sparse_penal": 0.01,
    "update_background": True,
    "normalize": True,
    "size_thres": (25, None),
}
param_first_temporal = {
    "noise_freq": 0.06,
    "sparse_penal": 1,
    "p": 1,
    "add_lag": 20,
    "jac_thres": 0.2,
    "zero_thres": 1e-8,
    "max_iters": 200,
    "use_smooth": True,
    "scs_fallback": False,
    "normalize": True,
    "post_scal": True,
}
param_first_merge = {"thres_corr": 0.8}
param_second_spatial = {
    "dl_wnd": 5,
    "sparse_penal": 0.01,
    "update_background": True,
    "normalize": True,
    "size_thres": (25, None),
}
param_second_temporal = {
    "noise_freq": 0.06,
    "sparse_penal": 1,
    "p": 1,
    "add_lag": 20,
    "jac_thres": 0.2,
    "zero_thres": 1e-8,
    "max_iters": 500,
    "use_smooth": True,
    "scs_fallback": False,
    "normalize": True,
    "post_scal": True,
}

os.environ["OMP_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["OPENBLAS_NUM_THREADS"] = "1"
os.environ["NUMBA_NUM_THREADS"] = "1"
os.environ["MINIAN_INTERMEDIATE"] = intpath

sys.path.append(minian_path)
from minian.cnmf import (
    compute_AtC,
    compute_trace,
    get_noise_fft,
    smooth_sig,
    unit_merge,
    update_spatial,
    update_temporal,
)
from minian.initialization import (
    gmm_refine,
    initA,
    initbf,
    initC,
    intensity_refine,
    ks_refine,
    pnr_refine,
    seeds_init,
    seeds_merge,
)
from minian.motion_correction import apply_shifts, estimate_shifts
from minian.preprocessing import denoise, remove_background
from minian.utilities import load_videos, open_minian, save_minian
from minian.visualization import (
    CNMFViewer,
    VArrayViewer,
    generate_videos,
    visualize_gmm_fit,
    visualize_preprocess,
    visualize_seeds,
    visualize_spatial_update,
    visualize_temporal_update,
    write_video,
)

dpath = os.path.abspath(dpath)
hv.notebook_extension("bokeh", width=100)

cluster = LocalCluster(n_workers=1, memory_limit="8GB")
client = Client(cluster)

varr = load_videos(dpath, **param_load_videos)

def test_pre_check ():
    dirpath = os.path.join(dpath, 'minian')
    if os.path.exists(dirpath) and os.path.isdir(dirpath):
        shutil.rmtree(dirpath)
    minianvideo = os.path.join(dpath, 'minian_mc.mp4')
    if os.path.exists(minianvideo):
        os.remove(minianvideo)
    assert os.path.exists(minianvideo) == False
    
def test_pipeline():
    global varr, subset_mc
    hv.output(size=output_size)
    if interactive:
        vaviewer = VArrayViewer(varr, framerate=5, summary=None)
        display(vaviewer.show())

    varr = save_minian(
        varr.rename("varr"),
        intpath,
        overwrite=True,
        chunks={"frame": 20, "height": -1, "width": -1},
    )

    if interactive:
        try:
            subset_mc = list(vaviewer.mask.values())[0]
        except IndexError:
            pass

    varr_ref = varr.sel(subset)

    varr_min = varr_ref.min("frame").compute()
    varr_ref = varr_ref - varr_min

    hv.output(size=output_size)
    if interactive:
        vaviewer = VArrayViewer(
            [varr.rename("original"), varr_ref.rename("glow_removed")],
            framerate=5,
            summary=None,
            layout=True,
        )
        display(vaviewer.show())

    hv.output(size=output_size)
    if interactive:
        display(
            visualize_preprocess(
                varr_ref.isel(frame=0).compute(),
                denoise,
                method=["median"],
                ksize=[5, 7, 9],
            )
        )

    varr_ref = denoise(varr_ref, **param_denoise)

    hv.output(size=output_size)
    if interactive:
        display(
            visualize_preprocess(
                varr_ref.isel(frame=0),
                remove_background,
                method=["tophat"],
                wnd=[10, 15, 20],
            )
        )

    varr_ref = remove_background(varr_ref, **param_background_removal)

    varr_ref = save_minian(varr_ref.rename("varr_ref"), dpath=intpath, overwrite=True)
    
    # Motion correction
    # def test_motion_correction():
    #     varr_ref = open_minian(intpath, "varr")
    shifts = estimate_shifts(varr_ref.sel(subset_mc), npart=34, **param_estimate_shift)

    shifts = save_minian(shifts.rename("shifts").chunk({"frame": 20}), **param_save_minian)

    test_shifts = open_minian(os.path.join(dpath_fixture, 'minian'),'shifts')
    assert shifts.all() == test_shifts.all(), "Test Fail: arrays are not the same";

    hv.output(size=output_size)
    if interactive:
        display(
            hv.NdOverlay(
                dict(
                    width=hv.Curve(shifts.sel(variable="width")),
                    height=hv.Curve(shifts.sel(variable="height")),
                )
            )
        )

    Y = apply_shifts(varr_ref, shifts)
    Y = Y.fillna(0)

    Y_fm_chk = save_minian(Y.astype(float).rename("Y_fm_chk"), intpath, overwrite=True)
    Y_hw_chk = save_minian(
        Y_fm_chk.rename("Y_hw_chk"),
        intpath,
        overwrite=True,
        chunks={"frame": -1, "height": 10, "width": 16},
    )

    hv.output(size=output_size)
    if interactive:
        vaviewer = VArrayViewer(
            [varr_ref.rename("before_mc"), Y_fm_chk.rename("after_mc")],
            framerate=5,
            summary=None,
            layout=True,
        )
        display(vaviewer.show())

    im_opts = dict(frame_width=500, aspect=752 / 480, cmap="Viridis", colorbar=True)
    (
        hv.Image(
            varr_ref.max("frame").compute(), ["width", "height"], label="before_mc"
        ).opts(**im_opts)
        + hv.Image(
            Y_hw_chk.max("frame").compute(), ["width", "height"], label="after_mc"
        ).opts(**im_opts)
    )

    vid_arr = xr.concat([varr_ref, Y_fm_chk], "width").chunk({"width": -1})
    write_video(vid_arr, "minian_mc.mp4", dpath)

    # Retrieve minian_mc.mp4 file inside fixture folder, used for testing
    fixture_probe = ffmpeg.probe(os.path.join(dpath_fixture, "minian_mc.mp4"))
    fixture_video_streams = [stream for stream in fixture_probe["streams"] if stream["codec_type"] == "video"]
        
    # Check 'minian_mc.mp4' was written to folder, with same size as the one in fixture folder
    assert os.path.exists(os.path.join(dpath, "minian_mc.mp4")) == True, "minian_mc.mp4 was written to local folder"
    assert os.path.getsize(os.path.join(dpath, "minian_mc.mp4")) == os.path.getsize(os.path.join(dpath_fixture, "minian_mc.mp4")) 
    
    probe = ffmpeg.probe(os.path.join(dpath, "minian_mc.mp4"))
    video_streams = [stream for stream in probe["streams"] if stream["codec_type"] == "video"]

    # Compare newly created minian_mc.mp4 file to the one in fixture folder
    assert video_streams[0]["width"] == fixture_video_streams[0]["width"]
    assert video_streams[0]["height"] == fixture_video_streams[0]["height"]
    assert video_streams[0]["codec_type"] == fixture_video_streams[0]["codec_type"]
    assert video_streams[0]["duration_ts"] == fixture_video_streams[0]["duration_ts"]
    assert video_streams[0]["codec_long_name"] == fixture_video_streams[0]["codec_long_name"]
   
    # Initialization
    # def test_initialization():
    #     Y_hw_chk = open_minian(intpath, "Y_hw_chk")
    #     Y_fm_chk = open_minian(intpath, "Y_fm_chk")
    max_proj = save_minian(
        Y_hw_chk.max("frame").rename("max_proj"), **param_save_minian
    ).compute()
    
    test_max_proj = open_minian(os.path.join(dpath_fixture, 'minian'),'max_proj')
    assert max_proj.all() == test_max_proj.all(), "Test Fail: arrays are not the same";

    seeds = seeds_init(Y_fm_chk, **param_seeds_init)
    
    hv.output(size=output_size)
    visualize_seeds(max_proj, seeds)

    if interactive:
        noise_freq_list = [0.005, 0.01, 0.02, 0.06, 0.1, 0.2, 0.3, 0.45, 0.6, 0.8]
        example_seeds = seeds.sample(6, axis="rows")
        example_trace = (
            Y_hw_chk.stack(spatial=["height", "width"])
            .sel(spatial=[tuple(hw) for hw in example_seeds[["height", "width"]].values])
            .assign_coords(spatial=np.arange(6))
            .rename(dict(spatial="seed"))
        )
        smooth_dict = dict()
        for freq in noise_freq_list:
            trace_smth_low = smooth_sig(example_trace, freq)
            trace_smth_high = smooth_sig(example_trace, freq, btype="high")
            trace_smth_low = trace_smth_low.compute()
            trace_smth_high = trace_smth_high.compute()
            hv_trace = hv.HoloMap(
                {
                    "signal": (
                        hv.Dataset(trace_smth_low)
                        .to(hv.Curve, kdims=["frame"])
                        .opts(frame_width=300, aspect=2, ylabel="Signal (A.U.)")
                    ),
                    "noise": (
                        hv.Dataset(trace_smth_high)
                        .to(hv.Curve, kdims=["frame"])
                        .opts(frame_width=300, aspect=2, ylabel="Signal (A.U.)")
                    ),
                },
                kdims="trace",
            ).collate()
            smooth_dict[freq] = hv_trace

    hv.output(size=output_size)
    if interactive:
        hv_res = (
            hv.HoloMap(smooth_dict, kdims=["noise_freq"])
            .collate()
            .opts(aspect=2)
            .overlay("trace")
            .layout("seed")
            .cols(3)
        )
        display(hv_res)

    seeds, pnr, gmm = pnr_refine(Y_hw_chk, seeds, **param_pnr_refine)

    if gmm:
        display(visualize_gmm_fit(pnr, gmm, 100))

    hv.output(size=output_size)
    visualize_seeds(max_proj, seeds, "mask_pnr")

    seeds = ks_refine(Y_hw_chk, seeds, **param_ks_refine)

    hv.output(size=output_size)
    visualize_seeds(max_proj, seeds, "mask_ks")

    seeds_final = seeds[seeds["mask_ks"] & seeds["mask_pnr"]].reset_index(drop=True)
    seeds_final = seeds_merge(Y_hw_chk, max_proj, seeds_final, **param_seeds_merge)

    hv.output(size=output_size)
    visualize_seeds(max_proj, seeds_final, "mask_mrg")

    A_init = initA(Y_hw_chk, seeds_final[seeds_final["mask_mrg"]], **param_initialize)
    A_init = save_minian(A_init.rename("A_init"), intpath, overwrite=True)

    C_init = initC(Y_fm_chk, A_init)
    C_init = save_minian(
        C_init.rename("C_init"), intpath, overwrite=True, chunks={"unit_id": 1, "frame": -1}
    )

    A, C = unit_merge(A_init, C_init, **param_first_merge)

    A = save_minian(A.rename("A"), intpath, overwrite=True)
    C = save_minian(C.rename("C"), intpath, overwrite=True)
    C_chk = save_minian(
        C.rename("C_chk"), intpath, overwrite=True, chunks={"unit_id": -1, "frame": 20}
    )

    b, f = initbf(Y_fm_chk, A, C_chk)
    b = save_minian(b.rename("b"), intpath, overwrite=True)
    f = save_minian(f.rename("f"), intpath, overwrite=True)
    
    im_opts = dict(
        frame_width=500,
        aspect=A.sizes["width"] / A.sizes["height"],
        cmap="Viridis",
        colorbar=True,
    )
    cr_opts = dict(frame_width=750, aspect=1.5 * A.sizes["width"] / A.sizes["height"])
    (
        hv.Image(A.sum("unit_id").rename("A").compute(), kdims=["width", "height"]).opts(
            **im_opts
        )
        + hv.Image(C.rename("C").compute(), kdims=["frame", "unit_id"]).opts(
            cmap="viridis", colorbar=True, **cr_opts
        )
        + hv.Image(b.rename("b").compute(), kdims=["width", "height"]).opts(**im_opts)
        + datashade(hv.Curve(f.rename("f").compute(), kdims=["frame"]), min_alpha=200).opts(
            **cr_opts
        )
    ).cols(2)

    # CNMF
    # def test_cnmf():
    try:
        client.close()
        cluster.close()
    except NameError:
        pass
    cluster = LocalCluster(
        n_workers=4, threads_per_worker=1, memory_limit="2GB", resources={"task": 1}
    )
    client = Client(cluster)

    # Y_hw_chk = open_minian(intpath, "Y_hw_chk")
    # A = open_minian(intpath, "A")
    # b = open_minian(intpath, "b")
    # C = open_minian(intpath, "C")
    # f = open_minian(intpath, "f")

    sn_spatial = get_noise_fft(Y_hw_chk, **param_get_noise)
    sn_spatial = save_minian(sn_spatial.rename("sn_spatial"), intpath, overwrite=True)

    if interactive:
        units = np.random.choice(A.coords["unit_id"], 10, replace=False)
        units.sort()
        A_sub = A.sel(unit_id=units).persist()
        C_sub = C.sel(unit_id=units).persist()

    if interactive:
        sprs_ls = [0.005, 0.01, 0.05]
        A_dict = dict()
        C_dict = dict()
        for cur_sprs in sprs_ls:
            cur_A, cur_b, cur_f, cur_mask = update_spatial(
                Y_hw_chk,
                A_sub,
                b,
                C_sub,
                f,
                sn_spatial,
                in_memory=True,
                dl_wnd=param_first_spatial["dl_wnd"],
                sparse_penal=cur_sprs,
            )
            if cur_A.sizes["unit_id"]:
                A_dict[cur_sprs] = cur_A.compute()
                C_dict[cur_sprs] = C_sub.sel(unit_id=cur_mask).compute()
        hv_res = visualize_spatial_update(A_dict, C_dict, kdims=["sparse penalty"])

    hv.output(size=output_size)
    if interactive:
        display(hv_res)

    A_new, b_new, f_new, mask = update_spatial(
        Y_hw_chk, A, b, C, f, sn_spatial, **param_first_spatial
    )

    hv.output(size=output_size)
    opts = dict(
        plot=dict(height=A.sizes["height"], width=A.sizes["width"], colorbar=True),
        style=dict(cmap="Viridis"),
    )
    (
        hv.Image(A.sum("unit_id").compute().rename("A"), kdims=["width", "height"])
        .opts(**opts)
        .relabel("Spatial Footprints Initial")
        + hv.Image(
            (A.fillna(0) > 0).sum("unit_id").compute().rename("A"),
            kdims=["width", "height"],
        )
        .opts(**opts)
        .relabel("Binary Spatial Footprints Initial")
        + hv.Image(A_new.sum("unit_id").compute().rename("A"), kdims=["width", "height"])
        .opts(**opts)
        .relabel("Spatial Footprints First Update")
        + hv.Image(
            (A_new > 0).sum("unit_id").compute().rename("A"),
            kdims=["width", "height"],
        )
        .opts(**opts)
        .relabel("Binary Spatial Footprints First Update")
    ).cols(2)

    hv.output(size=output_size)
    opts_im = dict(
        plot=dict(height=b.sizes["height"], width=b.sizes["width"], colorbar=True),
        style=dict(cmap="Viridis"),
    )
    opts_cr = dict(plot=dict(height=b.sizes["height"], width=b.sizes["height"] * 2))
    (
        hv.Image(b.compute(), kdims=["width", "height"])
        .opts(**opts_im)
        .relabel("Background Spatial Initial")
        + hv.Curve(f.compute(), kdims=["frame"])
        .opts(**opts_cr)
        .relabel("Background Temporal Initial")
        + hv.Image(b_new.compute(), kdims=["width", "height"])
        .opts(**opts_im)
        .relabel("Background Spatial First Update")
        + hv.Curve(f_new.compute(), kdims=["frame"])
        .opts(**opts_cr)
        .relabel("Background Temporal First Update")
    ).cols(2)

    A = save_minian(
        A_new.rename("A"),
        intpath,
        overwrite=True,
        chunks={"unit_id": 1, "height": -1, "width": -1},
    )
    b = save_minian(b_new.rename("b"), intpath, overwrite=True)
    f = save_minian(f_new.chunk({"frame": 20}).rename("f"), intpath, overwrite=True)
    C = C.sel(unit_id=A.coords["unit_id"].values)
    C_chk = C_chk.sel(unit_id=A.coords["unit_id"].values)

    if interactive:
        units = np.random.choice(A.coords["unit_id"], 10, replace=False)
        units.sort()
        A_sub = A.sel(unit_id=units).persist()
        C_sub = C_chk.sel(unit_id=units).persist()

    if interactive:
        p_ls = [1]
        sprs_ls = [0.1, 0.5, 1, 2]
        add_ls = [20]
        noise_ls = [0.06]
        YA_dict, C_dict, S_dict, g_dict, sig_dict, A_dict = [dict() for _ in range(6)]
        YrA = (
            compute_trace(Y_fm_chk, A_sub, b, C_sub, f)
            .persist()
            .chunk({"unit_id": 1, "frame": -1})
        )
        for cur_p, cur_sprs, cur_add, cur_noise in itt.product(
            p_ls, sprs_ls, add_ls, noise_ls
        ):
            ks = (cur_p, cur_sprs, cur_add, cur_noise)
            print(
                "p:{}, sparse penalty:{}, additional lag:{}, noise frequency:{}".format(
                    cur_p, cur_sprs, cur_add, cur_noise
                )
            )
            cur_C, cur_S, cur_b0, cur_c0, cur_g, cur_mask = update_temporal(
                A_sub,
                C_sub,
                YrA=YrA,
                sparse_penal=cur_sprs,
                p=cur_p,
                use_smooth=True,
                add_lag=cur_add,
                noise_freq=cur_noise,
            )
            YA_dict[ks], C_dict[ks], S_dict[ks], g_dict[ks], sig_dict[ks], A_dict[ks] = (
                YrA.compute(),
                cur_C.compute(),
                cur_S.compute(),
                cur_g.compute(),
                (cur_C + cur_b0 + cur_c0).compute(),
                A_sub.compute(),
            )
        hv_res = visualize_temporal_update(
            YA_dict,
            C_dict,
            S_dict,
            g_dict,
            sig_dict,
            A_dict,
            kdims=["p", "sparse penalty", "additional lag", "noise frequency"],
        )

    hv.output(size=output_size)
    if interactive:
        display(hv_res)

    YrA = save_minian(
        compute_trace(Y_fm_chk, A, b, C_chk, f).rename("YrA"), intpath, overwrite=True
    )

    C_new, S_new, b0_new, c0_new, g, mask = update_temporal(
        A, C, YrA=YrA, **param_first_temporal
    )

    hv.output(size=output_size)
    opts_im = dict(frame_width=500, aspect=2, colorbar=True, cmap="Viridis")
    (
        hv.Image(C.compute().rename("ci"), kdims=["frame", "unit_id"])
        .opts(**opts_im)
        .relabel("Temporal Trace Initial")
        + hv.Div("")
        + hv.Image(C_new.compute().rename("c1"), kdims=["frame", "unit_id"])
        .opts(**opts_im)
        .relabel("Temporal Trace First Update")
        + hv.Image(S_new.compute().rename("s1"), kdims=["frame", "unit_id"])
        .opts(**opts_im)
        .relabel("Spikes First Update")
    ).cols(2)

    hv.output(size=output_size)
    if interactive:
        h, w = A.sizes["height"], A.sizes["width"]
        im_opts = dict(aspect=w / h, frame_width=500, cmap="Viridis")
        cr_opts = dict(aspect=3, frame_width=1000)
        bad_units = mask.where(mask == False, drop=True).coords["unit_id"].values
        if len(bad_units) > 0:
            hv_res = (
                hv.NdLayout(
                    {
                        "Spatial Footprint": Dynamic(
                            hv.Dataset(A.sel(unit_id=bad_units).compute().rename("A"))
                            .to(hv.Image, kdims=["width", "height"])
                            .opts(**im_opts)
                        ),
                        "Spatial Footprints of Accepted Units": Dynamic(
                            hv.Image(
                                A.sel(unit_id=mask).sum("unit_id").compute().rename("A"),
                                kdims=["width", "height"],
                            ).opts(**im_opts)
                        ),
                    }
                )
                + datashade(
                    hv.Dataset(YrA.sel(unit_id=bad_units).rename("raw")).to(
                        hv.Curve, kdims=["frame"]
                    )
                )
                .opts(**cr_opts)
                .relabel("Temporal Trace")
            ).cols(1)
            display(hv_res)
        else:
            print("No rejected units to display")

    hv.output(size=output_size)
    if interactive:
        sig = C_new + b0_new + c0_new
        display(
            visualize_temporal_update(
                YrA.sel(unit_id=mask),
                C_new,
                S_new,
                g,
                sig,
                A.sel(unit_id=mask),
            )
        )

    C = save_minian(
        C_new.rename("C").chunk({"unit_id": 1, "frame": -1}), intpath, overwrite=True
    )
    C_chk = save_minian(
        C.rename("C_chk"), intpath, overwrite=True, chunks={"unit_id": -1, "frame": 20}
    )
    S = save_minian(
        S_new.rename("S").chunk({"unit_id": 1, "frame": -1}), intpath, overwrite=True
    )
    b0 = save_minian(
        b0_new.rename("b0").chunk({"unit_id": 1, "frame": -1}), intpath, overwrite=True
    )
    c0 = save_minian(
        c0_new.rename("c0").chunk({"unit_id": 1, "frame": -1}), intpath, overwrite=True
    )
    A = A.sel(unit_id=C.coords["unit_id"].values)

    A_mrg, C_mrg, [sig_mrg] = unit_merge(A, C, [C + b0 + c0], **param_first_merge)

    hv.output(size=output_size)
    opts_im = dict(frame_width=500, aspect=2, colorbar=True, cmap="Viridis")
    (
        hv.Image(C.compute().rename("c1"), kdims=["frame", "unit_id"])
        .relabel("Temporal Signals Before Merge")
        .opts(**opts_im)
        + hv.Image(C_mrg.compute().rename("c2"), kdims=["frame", "unit_id"])
        .relabel("Temporal Signals After Merge")
        .opts(**opts_im)
    )

    A = save_minian(A_mrg.rename("A_mrg"), intpath, overwrite=True)
    C = save_minian(C_mrg.rename("C_mrg"), intpath, overwrite=True)
    C_chk = save_minian(
        C.rename("C_mrg_chk"), intpath, overwrite=True, chunks={"unit_id": -1, "frame": 20}
    )
    sig = save_minian(sig_mrg.rename("sig_mrg"), intpath, overwrite=True)

    if interactive:
        units = np.random.choice(A.coords["unit_id"], 10, replace=False)
        units.sort()
        A_sub = A.sel(unit_id=units).persist()
        C_sub = sig.sel(unit_id=units).persist()

    if interactive:
        sprs_ls = [5e-3, 1e-2, 5e-2]
        A_dict = dict()
        C_dict = dict()
        for cur_sprs in sprs_ls:
            cur_A, cur_b, cur_f, cur_mask = update_spatial(
                Y_hw_chk,
                A_sub,
                b,
                C_sub,
                f,
                sn_spatial,
                in_memory=True,
                dl_wnd=param_second_spatial["dl_wnd"],
                sparse_penal=cur_sprs,
            )
            if cur_A.sizes["unit_id"]:
                A_dict[cur_sprs] = cur_A.compute()
                C_dict[cur_sprs] = C_sub.sel(unit_id=cur_mask).compute()
        hv_res = visualize_spatial_update(A_dict, C_dict, kdims=["sparse penalty"])

    hv.output(size=output_size)
    if interactive:
        display(hv_res)

    A_new, b_new, f_new, mask = update_spatial(
        Y_hw_chk, A, b, sig, f, sn_spatial, **param_second_spatial
    )

    hv.output(size=output_size)
    opts = dict(
        plot=dict(height=A.sizes["height"], width=A.sizes["width"], colorbar=True),
        style=dict(cmap="Viridis"),
    )
    (
        hv.Image(A.sum("unit_id").compute().rename("A"), kdims=["width", "height"])
        .opts(**opts)
        .relabel("Spatial Footprints Last")
        + hv.Image(
            (A.fillna(0) > 0).sum("unit_id").compute().rename("A"),
            kdims=["width", "height"],
        )
        .opts(**opts)
        .relabel("Binary Spatial Footprints Last")
        + hv.Image(A_new.sum("unit_id").compute().rename("A"), kdims=["width", "height"])
        .opts(**opts)
        .relabel("Spatial Footprints New")
        + hv.Image(
            (A_new > 0).sum("unit_id").compute().rename("A"),
            kdims=["width", "height"],
        )
        .opts(**opts)
        .relabel("Binary Spatial Footprints New")
    ).cols(2)

    hv.output(size=output_size)
    opts_im = dict(
        plot=dict(height=b.sizes["height"], width=b.sizes["width"], colorbar=True),
        style=dict(cmap="Viridis"),
    )
    opts_cr = dict(plot=dict(height=b.sizes["height"], width=b.sizes["height"] * 2))
    (
        hv.Image(b.compute(), kdims=["width", "height"])
        .opts(**opts_im)
        .relabel("Background Spatial Last")
        + hv.Curve(f.compute(), kdims=["frame"])
        .opts(**opts_cr)
        .relabel("Background Temporal Last")
        + hv.Image(b_new.compute(), kdims=["width", "height"])
        .opts(**opts_im)
        .relabel("Background Spatial New")
        + hv.Curve(f_new.compute(), kdims=["frame"])
        .opts(**opts_cr)
        .relabel("Background Temporal New")
    ).cols(2)

    A = save_minian(
        A_new.rename("A"),
        intpath,
        overwrite=True,
        chunks={"unit_id": 1, "height": -1, "width": -1},
    )
    b = save_minian(b_new.rename("b"), intpath, overwrite=True)
    f = save_minian(f_new.chunk({"frame": 20}).rename("f"), intpath, overwrite=True)
    C = C.sel(unit_id=A.coords["unit_id"].values)
    C_chk = C_chk.sel(unit_id=A.coords["unit_id"].values)

    if interactive:
        units = np.random.choice(A.coords["unit_id"], 10, replace=False)
        units.sort()
        A_sub = A.sel(unit_id=units).persist()
        C_sub = C_chk.sel(unit_id=units).persist()

    if interactive:
        p_ls = [1]
        sprs_ls = [0.1, 0.5, 1, 2]
        add_ls = [20]
        noise_ls = [0.06]
        YA_dict, C_dict, S_dict, g_dict, sig_dict, A_dict = [dict() for _ in range(6)]
        YrA = (
            compute_trace(Y_fm_chk, A_sub, b, C_sub, f)
            .persist()
            .chunk({"unit_id": 1, "frame": -1})
        )
        for cur_p, cur_sprs, cur_add, cur_noise in itt.product(
            p_ls, sprs_ls, add_ls, noise_ls
        ):
            ks = (cur_p, cur_sprs, cur_add, cur_noise)
            print(
                "p:{}, sparse penalty:{}, additional lag:{}, noise frequency:{}".format(
                    cur_p, cur_sprs, cur_add, cur_noise
                )
            )
            cur_C, cur_S, cur_b0, cur_c0, cur_g, cur_mask = update_temporal(
                A_sub,
                C_sub,
                YrA=YrA,
                sparse_penal=cur_sprs,
                p=cur_p,
                use_smooth=True,
                add_lag=cur_add,
                noise_freq=cur_noise,
            )
            YA_dict[ks], C_dict[ks], S_dict[ks], g_dict[ks], sig_dict[ks], A_dict[ks] = (
                YrA.compute(),
                cur_C.compute(),
                cur_S.compute(),
                cur_g.compute(),
                (cur_C + cur_b0 + cur_c0).compute(),
                A_sub.compute(),
            )
        hv_res = visualize_temporal_update(
            YA_dict,
            C_dict,
            S_dict,
            g_dict,
            sig_dict,
            A_dict,
            kdims=["p", "sparse penalty", "additional lag", "noise frequency"],
        )

    hv.output(size=output_size)
    if interactive:
        display(hv_res)

    YrA = save_minian(
        compute_trace(Y_fm_chk, A, b, C_chk, f).rename("YrA"), intpath, overwrite=True
    )

    C_new, S_new, b0_new, c0_new, g, mask = update_temporal(
        A, C, YrA=YrA, **param_first_temporal
    )

    hv.output(size=output_size)
    opts_im = dict(frame_width=500, aspect=2, colorbar=True, cmap="Viridis")
    (
        hv.Image(C.compute().rename("c1"), kdims=["frame", "unit_id"])
        .opts(**opts_im)
        .relabel("Temporal Trace Last")
        + hv.Image(S.compute().rename("s1"), kdims=["frame", "unit_id"])
        .opts(**opts_im)
        .relabel("Spikes Last")
        + hv.Image(C_new.compute().rename("c2"), kdims=["frame", "unit_id"])
        .opts(**opts_im)
        .relabel("Temporal Trace New")
        + hv.Image(S_new.compute().rename("s2"), kdims=["frame", "unit_id"])
        .opts(**opts_im)
        .relabel("Spikes New")
    ).cols(2)

    hv.output(size=output_size)
    if interactive:
        h, w = A.sizes["height"], A.sizes["width"]
        im_opts = dict(aspect=w / h, frame_width=500, cmap="Viridis")
        cr_opts = dict(aspect=3, frame_width=1000)
        bad_units = mask.where(mask == False, drop=True).coords["unit_id"].values
        if len(bad_units) > 0:
            hv_res = (
                hv.NdLayout(
                    {
                        "Spatial Footprint": Dynamic(
                            hv.Dataset(A.sel(unit_id=bad_units).compute().rename("A"))
                            .to(hv.Image, kdims=["width", "height"])
                            .opts(**im_opts)
                        ),
                        "Spatial Footprints of Accepted Units": Dynamic(
                            hv.Image(
                                A.sel(unit_id=mask).sum("unit_id").compute().rename("A"),
                                kdims=["width", "height"],
                            ).opts(**im_opts)
                        ),
                    }
                )
                + datashade(
                    hv.Dataset(YrA.sel(unit_id=bad_units).rename("raw")).to(
                        hv.Curve, kdims=["frame"]
                    )
                )
                .opts(**cr_opts)
                .relabel("Temporal Trace")
            ).cols(1)
            display(hv_res)
        else:
            print("No rejected units to display")

    hv.output(size=output_size)
    if interactive:
        sig = C_new + b0_new + c0_new
        display(
            visualize_temporal_update(
                YrA.sel(unit_id=mask),
                C_new,
                S_new,
                g,
                sig,
                A.sel(unit_id=mask),
            )
        )

    C = save_minian(
        C_new.rename("C").chunk({"unit_id": 1, "frame": -1}), intpath, overwrite=True
    )
    C_chk = save_minian(
        C.rename("C_chk"), intpath, overwrite=True, chunks={"unit_id": -1, "frame": 20}
    )
    S = save_minian(
        S_new.rename("S").chunk({"unit_id": 1, "frame": -1}), intpath, overwrite=True
    )
    b0 = save_minian(
        b0_new.rename("b0").chunk({"unit_id": 1, "frame": -1}), intpath, overwrite=True
    )
    c0 = save_minian(
        c0_new.rename("c0").chunk({"unit_id": 1, "frame": -1}), intpath, overwrite=True
    )
    A = A.sel(unit_id=C.coords["unit_id"].values)

    try:
        client.close()
        cluster.close()
    except NameError:
        pass
    cluster = LocalCluster(n_workers=1, memory_limit="8GB")
    client = Client(cluster)

    AC = save_minian(compute_AtC(A, C_chk).rename("AC"), intpath, overwrite=True)
    generate_videos(varr, Y_fm_chk, AC=AC, vpath=dpath)

    if interactive:
        cnmfviewer = CNMFViewer(A=A, C=C, S=S, org=Y_fm_chk)

    hv.output(size=output_size)
    if interactive:
        display(cnmfviewer.show())

    if interactive:
        A = A.assign_coords(unit_labels=("unit_id", cnmfviewer.unit_labels))
        C = C.assign_coords(unit_labels=("unit_id", cnmfviewer.unit_labels))
        S = S.assign_coords(unit_labels=("unit_id", cnmfviewer.unit_labels))
        c0 = c0.assign_coords(unit_labels=("unit_id", cnmfviewer.unit_labels))
        b0 = b0.assign_coords(unit_labels=("unit_id", cnmfviewer.unit_labels))

    test_A = open_minian(os.path.join(dpath_fixture, 'minian'),'A')
    assert A.all() == test_A.all(), "Test Fail: A does not match results in fixture folder";

    test_C = open_minian(os.path.join(dpath_fixture, 'minian'),'C')
    assert C.all() == test_C.all(), "Test Fail: C does not match results in fixture folder";

    test_S = open_minian(os.path.join(dpath_fixture, 'minian'),'S')
    assert S.all() == test_S.all(), "Test Fail: S does not match results in fixture folder";

    test_c0 = open_minian(os.path.join(dpath_fixture, 'minian'),'c0')
    assert c0.all() == test_c0.all(), "Test Fail: c0 does not match results in fixture folder";
    
    test_b0 = open_minian(os.path.join(dpath_fixture, 'minian'),'b0')
    assert b0.all() == test_b0.all(), "Test Fail: b0 does not match results in fixture folder";

    test_b = open_minian(os.path.join(dpath_fixture, 'minian'),'b')
    assert b.all() == test_b.all(), "Test Fail: b does not match results in fixture folder";
    
    test_f = open_minian(os.path.join(dpath_fixture, 'minian'),'f')
    assert f.all() == test_f.all(), "Test Fail: f does not match results in fixture folder";
    
    A = save_minian(A.rename("A"), **param_save_minian)
    C = save_minian(C.rename("C"), **param_save_minian)
    S = save_minian(S.rename("S"), **param_save_minian)
    c0 = save_minian(c0.rename("c0"), **param_save_minian)
    b0 = save_minian(b0.rename("b0"), **param_save_minian)
    b = save_minian(b.rename("b"), **param_save_minian)
    f = save_minian(f.rename("f"), **param_save_minian)


import pytest
import numpy as np
import holoviews as hv

from ..utilities import load_videos
from ..preprocessing import denoise, remove_background, stripe_correction

dpath = "./minian/test/test_movie"

param_load_videos = {
    "pattern": "msCam[0-9].avi",
    "dtype": np.uint8,
    "downsample": dict(frame=2, height=1, width=1),
    "downsample_strategy": "subset",
}

param_denoise = {"method": "median", "ksize": 7}

param_background_removal = {
    'method': 'tophat',
    'wnd': 15
}

@pytest.fixture
def varr():
    return load_videos(dpath, **param_load_videos)


def test_can_load_videos(varr):
    assert varr.shape[0] == 400  # frames
    assert varr.shape[1] == 480  # height
    assert varr.shape[2] == 752  # width
    return varr


def test_can_init_holoviews():
    hv.notebook_extension("bokeh")


def test_subset_part_video(varr):
    subset = dict(frame=slice(0, None))
    varr_ref = varr.sel(subset)
    assert varr_ref.all() == varr.all()

    
def test_remove_background(varr):
    varr_ref = denoise(varr, **param_denoise)
    varr_ref_remove = remove_background(varr_ref, **param_background_removal)
    assert (
        varr_ref.all() != varr_ref_remove.all()
    )  # when both are equal the denoise didn't do anything --> fail

def test_denoise(varr):
    varr_ref = denoise(varr, **param_denoise)
    assert (
        varr_ref.all() != varr.all()
    )  # when both are equal the denoise didn't do anything --> fail